from pydantic import BaseModel


class WhisperSentPackage(BaseModel):
    channel: str
    cmd: str
    userid_from: int
    text: str
    content: str  # processed by program
    time: int
    userid_to: int
    type: str = "whisper"



from pydantic import BaseModel
from typing import Optional, Literal, List, Union


class User(BaseModel):
    channel: str
    color: Union[str, bool]
    hash: str
    isBot: bool
    isme: bool
    level: int
    nick: str
    trip: Optional[str] = None
    uType: Literal["user", "mod", "admin"]
    userid: int


class OnlineSetPackage(BaseModel):
    cmd: str
    channel: str
    nicks: List[str]
    time: int
    users: List[User]



from pydantic import BaseModel
from typing import Literal


class InvitePackage(BaseModel):
    channel: str
    cmd: str
    from_nick: str  # do it in program
    inviteChannel: str
    text: str
    time: int
    to_userid: int
    type: Literal["invite"]



from pydantic import BaseModel
from typing import Union, Optional, Literal


class OnlineAddPackage(BaseModel):
    channel: str
    cmd: str
    color: Union[bool, str]
    hash: str
    isBot: bool
    level: int
    nick: str
    time: int
    trip: Optional[str] = None
    uType: Literal["user", "mod", "admin"]
    userid: int



from pydantic import BaseModel
from typing import Optional, Literal


class ChatPackage(BaseModel):
    cmd: str
    channel: str
    color: Optional[str] = None
    level: int
    nick: str
    text: str
    time: int
    trip: Optional[str] = None
    uType: Literal["user", "mod", "admin"]
    userid: int
    customId: Optional[str] = None



from typing import Optional, Literal, Callable, List, Dict, Any
from pydantic import BaseModel
from warnings import warn
from websocket import create_connection, WebSocket
from hvicorn.models.client import *
from hvicorn.models.server import *
from json import loads, dumps
from hvicorn.utils.generate_customid import generate_customid
from hvicorn.utils.json_to_object import json_to_object, verifyNick
from time import sleep
from traceback import format_exc
from logging import debug
from threading import Thread

WS_ADDRESS = "wss://hack.chat/chat-ws"


def threaded(func):
    def wrapper(*args, **kwargs):
        Thread(target=func, args=tuple(args), kwargs=kwargs).start()

    return wrapper


class CommandContext:
    def __init__(
        self,
        bot: "Bot",
        sender: User,
        triggered_via: Literal["chat", "whisper"],
        text: str,
    ) -> None:
        self.bot = bot
        self.sender = sender
        self.triggered_via = triggered_via
        self.text = text

    def respond(self, text, at_sender=True):
        if self.triggered_via == "chat":
            self.bot.send_message(
                ("@" + self.sender.nick + " " if at_sender else "") + str(text)
            )
        elif self.triggered_via == "whisper":
            self.bot.whisper(self.sender.nick, text)
        else:
            warn("Unknown trigger method, ignoring")


class Bot:
    def __init__(self, nick: str, channel: str, password: Optional[str] = None) -> None:
        self.nick = nick
        self.channel = channel
        self.password = password
        self.websocket: Optional[WebSocket] = None
        self.startup_functions: List[Callable] = []
        self.event_functions: Dict[Any, List[Callable]] = {
            "__GLOBAL__": [self._internal_handler]
        }
        self.killed: bool = False
        self.users: List[User] = []
        self.commands: Dict[str, Callable] = {}

    def _send_model(self, model: BaseModel) -> None:
        try:
            data = model.model_dump()
        except:
            warn(f"Cannot stringify model, ignoring: {model}")
        payload = {}
        for k, v in data.items():
            if v != None:
                payload.update({k: v})
        if self.websocket:
            debug(f"Sent payload: {payload}")
            self.websocket.send(dumps(payload))
        else:
            warn(f"Websocket isn't open, ignoring: {model}")

    def get_user(
        self,
        by: Literal[
            "nick", "hash", "trip", "color", "isBot", "level", "uType", "userid"
        ],
        matches: str,
    ) -> Optional[User]:
        for user in self.users:
            if user.__dict__.get(by) == matches:
                return user
        return None

    def get_users(
        self,
        by: Literal[
            "nick", "hash", "trip", "color", "isBot", "level", "uType", "userid"
        ],
        matches: str,
    ) -> List[User]:
        results = []
        for user in self.users:
            if user.__dict__.get(by) == matches:
                results.append(user)
        return results

    def get_user_by_nick(self, nick: str) -> Optional[User]:
        return self.get_user("nick", nick)

    def _internal_handler(self, bot: "Bot", event: BaseModel) -> None:
        if isinstance(event, OnlineSetPackage):
            self.users = event.users
        elif isinstance(event, OnlineAddPackage):
            self.users.append(
                User(
                    channel=event.channel,
                    color=event.color,
                    hash=event.hash,
                    isBot=event.isBot,
                    isme=False,
                    level=event.level,
                    nick=event.nick,
                    trip=event.trip,
                    uType=event.uType,
                    userid=event.userid,
                )
            )
        elif isinstance(event, OnlineRemovePackage):
            if self.get_user_by_nick(event.nick):
                self.users.remove(self.get_user_by_nick(event.nick))
        if isinstance(event, ChatPackage):
            for command in self.commands.items():
                if event.text.startswith(command[0]):
                    try:
                        command[1](
                            CommandContext(
                                self,
                                self.get_user_by_nick(event.nick),
                                "chat",
                                event.text,
                            )
                        )
                    except:
                        warn(f"Ignoring exception in command: \n{format_exc()}")
        if isinstance(event, WhisperPackage):
            for command in self.commands.items():
                if event.content.startswith(command[0]):
                    try:
                        command[1](
                            CommandContext(
                                self,
                                self.get_user_by_nick(event.nick),
                                "whisper",
                                event.content,
                            )
                        )
                    except:
                        warn(f"Ignoring exception in command: \n{format_exc()}")

    def _connect(self) -> None:
        self.websocket = create_connection(WS_ADDRESS)
        while not self.websocket.connected:
            sleep(1)

    def _run_events(self, event_type: Any, args: list):
        for function in self.event_functions.get(event_type, []):
            try:
                function(*args)
            except:
                warn(f"Ignoring exception in event: \n{format_exc()}")

    def join(self) -> None:
        self._send_model(
            JoinRequest(nick=self.nick, channel=self.channel, password=self.password)
        )
        sleep(1)

    def send_message(self, text, editable=False) -> Message:
        customId = generate_customid() if editable else None
        self._send_model(ChatRequest(text=text, customId=customId))

        def wrapper(
            mode: Literal["overwrite", "prepend", "append", "complete"], text: str
        ):
            self._send_model(
                UpdateMessageRequest(customId=msg.customId, mode=mode, text=text)
            )

        msg = Message(text, customId)
        msg._edit = wrapper
        return msg

    def whisper(self, nick: str, text: str) -> None:
        self._send_model(WhisperRequest(nick=nick, text=text))

    def emote(self, text: str) -> None:
        self._send_model(EmoteRequest(text=text))

    def change_color(self, color: str = "reset") -> None:
        self._send_model(ChangeColorRequest(color=color))

    def change_nick(self, nick: str) -> None:
        if not verifyNick(nick):
            raise ValueError("Invaild Nickname")
        self._send_model(ChangeNickRequest(nick=nick))

    def invite(self, nick: str, channel: Optional[str] = None) -> None:
        self._send_model(InviteRequest(nick=nick, to=channel))

    def on(self, event_type: Any = None) -> None:
        def wrapper(func: Callable):
            nonlocal event_type
            if event_type is None:
                event_type = "__GLOBAL__"
            if event_type in self.event_functions.keys():
                self.event_functions[event_type].append(func)
                debug(f"Added handler for {event_type}: {func}")
            else:
                self.event_functions[event_type] = [func]
                debug(f"Set handler for {event_type} to {func}")

        return wrapper

    def startup(self, function: Callable) -> None:
        self.startup_functions.append(function)
        debug(f"Added startup function: {function}")
        return None

    def command(self, prefix: str) -> None:
        def wrapper(func: Callable):
            self.commands[prefix] = func

        return wrapper

    def kill(self) -> None:
        self.killed = True
        debug("Killing ws")
        self.websocket.close()
    
    def load_plugin(self, plugin_name: str, *args, **kwargs) -> None:
        try:
            plugin = __import__(plugin_name)
        except ImportError:
            warn(f"Failed to load plugin {plugin_name}, ignoring")
            return
        if "plugin_init" not in dir(plugin):
            warn(f"Failed to find init function of plugin {plugin_name}, ignoring")
            return
        if not callable(plugin.plugin_init):
            warn(f"Init function of plugin {plugin_name} isn't callable, ignoring")
            return
        try:
            plugin.plugin_init(self, *args, **kwargs)
        except:
            warn(f"Failed to init plugin {plugin_name}: \n{format_exc()}")
            return
        debug(f"Loaded plugin {plugin_name}")

    def run(self, ignore_self: bool = True) -> None:
        self._connect()
        self.join()
        for function in self.startup_functions:
            debug(f"Running startup function: {function}")
            function(self)
        while not self.killed:
            package = self.websocket.recv()
            if not package:
                debug("Killed")
                self.killed = True
                break
            package = loads(package)
            try:
                event = json_to_object(package)
            except Exception as e:
                debug(e)
                warn(
                    f"Failed to parse event, ignoring: {package} cause exception: \n{format_exc()}"
                )
                continue
            debug(f"Got event {type(event)}::{str(event)}")
            try:
                if event.nick == self.nick and ignore_self:
                    debug("Found self.nick, ignoring")
                    continue
            except:
                debug("No nick provided in event, passing loopcheck")
            self._run_events("__GLOBAL__", [self, event])
            self._run_events(type(event), [self, event])



from hvicorn.models.client.chat import ChatRequest, Message
from hvicorn.models.client.join import JoinRequest
from hvicorn.models.client.update_message import UpdateMessageRequest
from hvicorn.models.client.whisper import WhisperRequest
from hvicorn.models.client.emote import EmoteRequest
from hvicorn.models.client.changecolor import ChangeColorRequest
from hvicorn.models.client.changenick import ChangeNickRequest
from hvicorn.models.client.invite import InviteRequest



from pydantic import BaseModel
from typing import Union, Optional, Literal


class UpdateUserPackage(BaseModel):
    channel: Optional[str] = None
    cmd: str
    color: Union[bool, str]
    hash: Optional[str] = None
    isBot: Optional[bool] = None
    level: Optional[int] = None
    nick: Optional[str] = None
    time: Optional[int] = None
    trip: Optional[str] = None
    uType: Optional[Literal["user", "mod", "admin"]] = None
    userid: Optional[int] = None



from pydantic import BaseModel
from typing import Optional, Literal


class WhisperPackage(BaseModel):
    channel: str
    cmd: str
    nick: str  # didn't use from because it's reserved py keyword
    level: int
    text: str
    content: str  # process it in the program!!!
    time: int
    userid_to: int
    trip: Optional[str] = None
    type: Literal["whisper"]
    uType: Literal["user", "mod", "admin"]



from pydantic import BaseModel
from typing import Optional, Literal


class EmotePackage(BaseModel):
    channel: str
    cmd: Literal["emote"]
    nick: str
    text: str
    content: str  # do this in the program.
    time: int
    trip: Optional[str] = None
    userid: int



from pydantic import BaseModel
from typing import Optional, Literal
from hvicorn.models.client.update_message import UpdateMessageRequest


class Message:
    def __init__(self, text: str, customId: Optional[str] = None) -> None:
        self.text = text
        self.customId = customId
        self.editable = customId != None

    def _edit(
        self, mode: Literal["overwrite", "prepend", "append", "complete"], text: str
    ):
        return UpdateMessageRequest(customId=self.customId, mode=mode, text=text)

    def edit(self, text):
        return self._edit("overwrite", text)

    def prepend(self, text):
        return self._edit("prepend", text)

    def append(self, text):
        return self._edit("append", text)

    def complete(self):
        self.editable = False
        return UpdateMessageRequest(customId=self.customId, mode="complete")


class ChatRequest(BaseModel):
    cmd: str = "chat"
    text: str
    customId: Optional[str] = None



from setuptools import setup, find_packages

setup(
    name='hvicorn',
    version="0.0.3", 
    description='Something for creating webapps like uvicorn, But for online chatrooms...', 
    long_description=open('README.md',"r").read(), 
    long_description_content_type='text/markdown', 
    url='https://github.com/Hiyoteam/hvicorn', 
    author='0x24a', 
    author_email='admin@thz.cool', 
    license='MIT', 
    classifiers=[ 
        'Development Status :: 3 - Alpha', 
        'Intended Audience :: Developers', 
        'Topic :: Software Development :: Libraries :: Python Modules', 
        'Programming Language :: Python :: 3', 
        'Programming Language :: Python :: 3.12', 
    ],
    keywords='bot hack.chat hack.chat-bot', 
    packages=find_packages('.', include=["hvicorn", "hvicorn.*"]), 
    install_requires=["websocket-client==1.8.0","pydantic==2.7.3","setuptools==70.0.0"]
)


from hvicorn.models.server import (
    ChatPackage,
    EmotePackage,
    InfoPackage,
    InvitePackage,
    OnlineAddPackage,
    OnlineRemovePackage,
    OnlineSetPackage,
    UpdateUserPackage,
    WarnPackage,
    WhisperPackage,
    UpdateMessagePackage,
    ChangeNickPackage,
    CaptchaPackage,
    LockroomPackage,
    WhisperSentPackage,
)
from typing import Union
import string


def verifyNick(nick: str) -> bool:
    if len(nick) > 24:
        return False
    for char in nick:
        if char not in string.ascii_letters + string.digits + "_":
            return False
    return True


def json_to_object(
    data: dict,
) -> Union[
    ChatPackage,
    EmotePackage,
    InfoPackage,
    InvitePackage,
    OnlineAddPackage,
    OnlineRemovePackage,
    OnlineSetPackage,
    UpdateUserPackage,
    WarnPackage,
    WhisperPackage,
    UpdateMessagePackage,
    ChangeNickPackage,
    CaptchaPackage,
    LockroomPackage,
    WhisperSentPackage,
]:
    if not data.get("cmd"):
        raise ValueError("No `cmd` provided")
    command = data.get("cmd")
    if command == "chat":
        return ChatPackage(**data)
    elif command == "emote":
        data["content"] = data["text"].split(" ", 1)[1]
        return EmotePackage(**data)
    elif command == "info":
        if not data.get("type"):
            if (
                data.get("text", "").count(" ") == 3
                and data.get("text", "").split(" ", 1)[1].startswith("is now")
                and verifyNick(data.get("text", "").split()[0])
                and verifyNick(data.get("text", "").split()[3])
            ):
                data["old_nick"] = data.get("text", "").split(" ")[0]
                data["new_nick"] = data.get("text", "").split(" ")[3]
                return ChangeNickPackage(**data)
            elif (
                data.get("text", "")
                == "You have been denied access to that channel and have been moved somewhere else. Retry later or wait for a mod to move you."
            ):
                return LockroomPackage(time=data["time"])
            return InfoPackage(**data)
        elif data.get("type") == "whisper":
            if data.get("text", "").startswith("You whispered to"):
                data["userid_from"] = data["from"]
                data["userid_to"] = data["to"]
                del data["from"]
                del data["to"]
                data["content"] = data["text"].split(": ", 1)[1]
                return WhisperSentPackage(**data)
            else:
                data["nick"] = data["from"]
                data["userid_to"] = data["to"]
                del data["from"]
                del data["to"]
                data["content"] = data["text"].split(": ", 1)[1]
                return WhisperPackage(**data)
        elif data.get("type") == "invite":
            data["from_nick"] = data["from"]
            data["to_userid"] = data["to"]
            del data["from"]
            del data["to"]
            return InvitePackage(**data)
    elif command == "onlineSet":
        return OnlineSetPackage(**data)
    elif command == "onlineAdd":
        return OnlineAddPackage(**data)
    elif command == "onlineRemove":
        return OnlineRemovePackage(**data)
    elif command == "updateUser":
        return UpdateUserPackage(**data)
    elif command == "updateMessage":
        return UpdateMessagePackage(**data)
    elif command == "captcha":
        return CaptchaPackage(**data)
    elif command == "warn":
        return WarnPackage(**data)
    else:
        raise ValueError("Unknown package type")
